from array import array
from threading import Condition

class SendBuffer(object):

    def __init__(self, max_size):
        self.mark = 0
        self.buf = array('c')
        self.max_size = max_size
        self.full = Condition()

    def __len__(self):
        with self.full:
            return len(self.buf) - self.mark


    def write(self, data):
        with self.full:
            while len(self) + len(data) > self.max_size:
                # wait until data is written
                self.full.wait()

            self.buf.fromstring(data)

    def peek(self):
        with self.full:
            return memoryview(self.buf)[self.mark:]

    def written(self, n):
        with self.full:
            self.mark += n
            self.full.notify_all()

            if self.mark >= len(self.buf):
                self.mark = 0
                self.buf = array('c')
            elif self.mark >= self.max_size:
                self.buf = self.buf[self.mark:]
                self.mark = 0



from cffi import FFI

ffi = FFI()
ffi.cdef("""
    typedef struct redisReply {
        int type; /* REDIS_REPLY_* */
        long long integer; /* The integer when type is REDIS_REPLY_INTEGER */
        int len; /* Length of string */
        char *str; /* Used for both REDIS_REPLY_ERROR and REDIS_REPLY_STRING */
        size_t elements; /* number of elements, for REDIS_REPLY_ARRAY */
        struct redisReply **element; /* elements vector for REDIS_REPLY_ARRAY */
    } redisReply;

    struct redisReader {
        ...;
    };

    struct redisReader *redisReaderCreate(void);
    void redisReaderFree(struct redisReader *r);
    int redisReaderFeed(struct redisReader *r, const char *buf, size_t len);
    int redisReaderGetReply(struct redisReader *r, redisReply **reply);
    void freeReplyObject(void *reply);
""")

hiredis = ffi.verify("""
    #include "hiredis.h"
    """,
    sources=[
        "hiredis/hiredis.c",
        "hiredis/net.c",
        "hiredis/sds.c",
    ],
    include_dirs=["hiredis"])

class HiredisError(Exception):
    pass

class RedisReader(object):
    "Hiredis wrapper"

    REDIS_ERR = -1
    REDIS_OK = 0

    REDIS_REPLY_STRING = 1
    REDIS_REPLY_ARRAY = 2
    REDIS_REPLY_INTEGER = 3
    REDIS_REPLY_NIL = 4
    REDIS_REPLY_STATUS = 5
    REDIS_REPLY_ERROR = 6

    def __init__(self):
        self.hiredis = ffi.gc(
            hiredis.redisReaderCreate(),
            hiredis.redisReaderFree)

    def to_list(self, reply):
        res = []
        for idx in xrange(reply.elements):
            element = reply.element[idx]
            res.append(self.to_py(element))

        return res

    def to_py(self, reply):
        if reply.type == self.REDIS_REPLY_STRING:
            return ffi.buffer(reply.str, reply.len)[:]
        elif reply.type == self.REDIS_REPLY_ARRAY:
            return self.to_list(reply)
        elif reply.type == self.REDIS_REPLY_INTEGER:
            return reply.integer
        elif reply.type == self.REDIS_REPLY_NIL:
            return None
        elif reply.type == self.REDIS_REPLY_STATUS:
            return ffi.buffer(reply.str, reply.len)[:]
        elif reply.type == self.REDIS_REPLY_ERROR:
            raise HiredisError(ffi.buffer(reply.str, reply.len)[:])

    def feed(self, data):
        res = hiredis.redisReaderFeed(self.hiredis, data, len(data))
        if res == self.REDIS_ERR:
            raise HiredisError("Error reading")

    def get_reply(self):
        reply_pointer = ffi.new("redisReply * *")
        res = hiredis.redisReaderGetReply(self.hiredis, reply_pointer)
        if res == self.REDIS_ERR:
            raise RedisError("Error parsing")

        reply = reply_pointer[0]
        if reply:
            gc_reply = ffi.gc(reply, hiredis.freeReplyObject)
            return self.to_py(gc_reply)
        else:
            return False




from __future__ import absolute_import
from pypredis.future import Future
from pypredis.reader import RedisReader
from pypredis.sendbuffer import SendBuffer
from select import poll, POLLIN, POLLPRI, POLLOUT, POLLERR, POLLHUP, POLLNVAL
from Queue import Queue, Empty
from collections import defaultdict, namedtuple, deque
from threading import Thread
import socket
import os

#debug
from time import sleep

def pack_command(args):
    "Pack a series of arguments into a value Redis command"
    args_output = ''.join([
        ''.join(('$', str(len(k)), '\r\n', k, '\r\n'))
        for k in args])
    output = ''.join(
        ('*', str(len(args)), '\r\n', args_output))
    return output

def drain(q):
    while True:
        try:
            yield q.get_nowait()
        except Empty:
            break

RedisCommand = namedtuple("RedisCommand", ["result", "connection", "command"])

class BaseConnection(object):

    def __init__(self, buf_size=1000000, **connect_params):
        self.sock = None
        self.buf = SendBuffer(buf_size)
        self.resq = deque()
        self.reader = RedisReader()
        self.pid = os.getpid()
        self.connect_params = connect_params
        self.connect(**connect_params)

    @property
    def fd(self):
        return self.sock.fileno()

    @property
    def waiting(self):
        return bool(self.resq)

    def checkpid(self):
        if self.pid != os.getpid():
            self.disconnect()
            self.connect(**self.connect_params)

    def disconnect(self):
        self.sock.close()

    def write(self, cmd):
        self.resq.append(cmd.result)
        self.buf.write(cmd.command)

    def pump_out(self):
        try:
            data = self.buf.peek()
            if data:
                n = self.sock.send(data)
                self.buf.written(n)
        except Exception as e:
            res = self.resq.popleft()
            res.set_exception(e)

    
    def pump_in(self):
        try:
            data = self.sock.recv(4096)
            self.reader.feed(data)
            while True:
                reply = self.reader.get_reply()
                if reply == False:
                    break
                res = self.resq.popleft()
                res.set_result(reply)
        except Exception as e:
            res = self.resq.popleft()
            res.set_exception(e)

class UnixConnection(BaseConnection):

    def connect(self, path):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(path)
        self.sock.setblocking(0)


class EventLoop(Thread):
    
    def __init__(self):
        Thread.__init__(self)
        self.daemon = True
        self.queue = Queue()
        self.timeout = 0.1
        self.poll = poll()
        self.fd_index = {}

    def send_command(self, conn, *args):
        cmdstr = pack_command(args)
        res = Future()
        cmd = RedisCommand(res, conn, cmdstr)
        conn.write(cmd)
        self.queue.put(cmd)
        return res

    def _register(self, conn):
        if conn.fd not in self.fd_index:
            conn.checkpid()
            self.poll.register(conn.fd)
            self.fd_index[conn.fd] = conn

    def _unregister(self, fd):
        self.poll.unregister(fd)
        del self.fd_index[fd]

    def _handle_events(self, events):
        for fd, e in events:
            if e & POLLOUT:
                self._handle_writes(fd)
            if e & (POLLIN | POLLPRI):
                self._handle_reads(fd)

    def _handle_writes(self, fd):
        conn = self.fd_index[fd]
        conn.pump_out()

    def _handle_reads(self, fd):
        conn = self.fd_index[fd]
        conn.pump_in()
        if not conn.waiting:
            self._unregister(fd)

    def run(self):
        while True:
            commands = drain(self.queue)
            for cmd in commands:
                self._register(cmd.connection)

            events = self.poll.poll(self.timeout)
            self._handle_events(events)



from distutils.core import setup
from distutils.command.build import build
from distutils.command.install import install

class CFFIBuild(build):

    def finalize_options(self):
        import pypredis.reader
        self.distribution.ext_modules = [pypredis.reader.ffi.verifier.get_extension()]
        build.finalize_options(self)


class CFFIInstall(install):

    def finalize_options(self):
        import pypredis.reader
        self.distribution.ext_modules = [pypredis.reader.ffi.verifier.get_extension()]
        install.finalize_options(self)

setup(name='pypredis',
      version='0.2',
      description='A heavy-duty Redis client',
      author='Pepij de Vos',
      author_email='pepijndevos@gmail.com',
      url='https://github.com/pepijndevos/pypredis',
      install_requires=['cffi'],
      setup_requires=['cffi'],
      packages=['pypredis'],
      data_files=[('hiredis', ['hiredis/hiredis.h', 'hiredis/fmacros.h', 'hiredis/net.h', 'hiredis/sds.h', ])],
      zip_safe=False,
      cmdclass={
          "build": CFFIBuild,
          "install": CFFIInstall,
      },
     )



import threading

class TimeoutError(Exception):
    pass

class Future(object):
    "Backport of concurrent.futures.Future"

    def __init__(self):
        self.event = threading.Event()
        self.callbacks = []
        self.lock = threading.Lock()
        self._result = None
        self._exception = None

    def done(self):
        return self.event.is_set()

    def result(self, timeout=None):
        if self.event.wait(timeout):
            if self._exception:
                raise self._exception
            else:
                return self._result
        else:
            raise TimeoutError()

    def exception(self, timeout=None):
        if self.event.wait(timeout):
            return self._exception
        else:
            raise TimeoutError()

    def add_done_callback(self, cb):
        with self.lock:
            if self.event.is_set():
                cb(self)
            else:
                self.callbacks.append(cb)

    def set_result(self, result):
        with self.lock:
            self._result = result
            self.event.set()
            for cb in self.callbacks:
                cb(self)

    def set_exception(self, exception):
        with self.lock:
            self._exception = exception
            self.event.set()
            for cb in self.callbacks:
                cb(self)
